@@ -2,8 +2,12 @@ package Module::CPANfile;
 use strict;
 use warnings;
 use Cwd;
+use Carp ();
+use Module::CPANfile::Environment;
+use Module::CPANfile::Features;
+use Module::CPANfile::Requirement;
 
-our $VERSION = '0.9023';
+our $VERSION = '1.0900';
 
 sub new {
     my($class, $file) = @_;
@@ -12,9 +16,9 @@ sub new {
 
 sub load {
     my($proto, $file) = @_;
+
     my $self = ref $proto ? $proto : $proto->new;
-    $self->{file} = $file || "cpanfile";
-    $self->parse;
+    $self->parse($file || Cwd::abs_path('cpanfile'));
     $self;
 }
 
@@ -26,32 +30,73 @@ sub save {
 }
 
 sub parse {
-    my $self = shift;
+    my($self, $file) = @_;
 
-    my $file = Cwd::abs_path($self->{file});
-    $self->{result} = Module::CPANfile::Environment::parse($file) or die $@;
+    my $code = do {
+        open my $fh, "<", $file or die "$file: $!";
+        join '', <$fh>;
+    };
+
+    my $env = Module::CPANfile::Environment->new($file);
+    $env->parse($code) or die $@;
+
+    $self->{_prereqs} = $env->prereqs;
 }
 
 sub from_prereqs {
     my($proto, $prereqs) = @_;
 
     my $self = $proto->new;
-    $self->{result} = Module::CPANfile::Result->from_prereqs($prereqs);
+    $self->{_prereqs} = Module::CPANfile::Prereqs->from_cpan_meta($prereqs);
 
     $self;
 }
 
-sub prereqs { shift->prereq }
-
-sub prereq {
+sub features {
     my $self = shift;
-    require CPAN::Meta::Prereqs;
-    CPAN::Meta::Prereqs->new($self->prereq_specs);
+    map $self->feature($_), $self->{_prereqs}->identifiers;
+}
+
+sub feature {
+    my($self, $identifier) = @_;
+    $self->{_prereqs}->feature($identifier);
+}
+
+sub prereq { shift->prereqs }
+
+sub prereqs {
+    my $self = shift;
+    $self->{_prereqs}->as_cpan_meta;
+}
+
+sub effective_prereqs {
+    my($self, $features) = @_;
+    $self->prereqs_with(@{$features || []});
+}
+
+sub prereqs_with {
+    my($self, @feature_identifiers) = @_;
+
+    my $prereqs = $self->prereqs;
+    my @others = map { $self->feature($_)->prereqs } @feature_identifiers;
+
+    $prereqs->with_merged_prereqs(\@others);
 }
 
 sub prereq_specs {
     my $self = shift;
-    $self->{result}{spec};
+    $self->prereqs->as_string_hash;
+}
+
+sub prereq_for_module {
+    my($self, $module) = @_;
+    $self->{_prereqs}->find($module);
+}
+
+sub options_for_module {
+    my($self, $module) = @_;
+    my $prereq = $self->prereq_for_module($module) or return;
+    $prereq->requirement->options;
 }
 
 sub merge_meta {
@@ -70,14 +115,38 @@ sub merge_meta {
     CPAN::Meta->new($struct)->save($file, { version => $version });
 }
 
+sub _dump {
+    my $str = shift;
+    require Data::Dumper;
+    chomp(my $value = Data::Dumper->new([$str])->Terse(1)->Dump);
+    $value;
+}
+
 sub to_string {
     my($self, $include_empty) = @_;
 
-    my $prereqs = $self->{result}{spec};
+    my $prereqs = $self->prereq_specs;
+
+    my $code = '';
+    $code .= $self->_dump_prereqs($prereqs, $include_empty);
+
+    for my $feature ($self->features) {
+        $code .= sprintf "feature %s, %s => sub {\n", _dump($feature->{identifier}), _dump($feature->{description});
+        $code .= $self->_dump_prereqs($feature->{spec}, $include_empty, 4);
+        $code .= "}\n\n";
+    }
+
+    $code =~ s/\n+$/\n/s;
+    $code;
+}
+
+sub _dump_prereqs {
+    my($self, $prereqs, $include_empty, $base_indent) = @_;
 
     my $code = '';
     for my $phase (qw(runtime configure build test develop)) {
         my $indent = $phase eq 'runtime' ? '' : '    ';
+        $indent = (' ' x ($base_indent || 0)) . $indent;
 
         my($phase_code, $requirements);
         $phase_code .= "on $phase => sub {\n" unless $phase eq 'runtime';
@@ -102,128 +171,6 @@ sub to_string {
     $code;
 }
 
-package Module::CPANfile::Environment;
-use strict;
-
-my @bindings = qw(
-    on requires recommends suggests conflicts
-    osname
-    configure_requires build_requires test_requires author_requires
-);
-
-my $file_id = 1;
-
-sub import {
-    my($class, $result_ref) = @_;
-    my $pkg = caller;
-
-    $$result_ref = Module::CPANfile::Result->new;
-    for my $binding (@bindings) {
-        no strict 'refs';
-        *{"$pkg\::$binding"} = sub { $$result_ref->$binding(@_) };
-    }
-}
-
-sub parse {
-    my $file = shift;
-
-    my $code = do {
-        open my $fh, "<", $file or die "$file: $!";
-        join '', <$fh>;
-    };
-
-    my($res, $err);
-
-    {
-        local $@;
-        $res = eval sprintf <<EVAL, $file_id++;
-package Module::CPANfile::Sandbox%d;
-no warnings;
-my \$_result;
-BEGIN { import Module::CPANfile::Environment \\\$_result };
-
-$code;
-
-\$_result;
-EVAL
-        $err = $@;
-    }
-
-    if ($err) { die "Parsing $file failed: $err" };
-
-    return $res;
-}
-
-package Module::CPANfile::Result;
-use strict;
-
-sub from_prereqs {
-    my($class, $spec) = @_;
-    bless {
-        phase => 'runtime',
-        spec => $spec,
-    }, $class;
-}
-
-sub new {
-    bless {
-        phase => 'runtime', # default phase
-        spec  => {},
-    }, shift;
-}
-
-sub on {
-    my($self, $phase, $code) = @_;
-    local $self->{phase} = $phase;
-    $code->()
-}
-
-sub osname { die "TODO" }
-
-sub requires {
-    my($self, $module, $requirement) = @_;
-    $self->{spec}{$self->{phase}}{requires}{$module} = $requirement || 0;
-}
-
-sub recommends {
-    my($self, $module, $requirement) = @_;
-    $self->{spec}->{$self->{phase}}{recommends}{$module} = $requirement || 0;
-}
-
-sub suggests {
-    my($self, $module, $requirement) = @_;
-    $self->{spec}->{$self->{phase}}{suggests}{$module} = $requirement || 0;
-}
-
-sub conflicts {
-    my($self, $module, $requirement) = @_;
-    $self->{spec}->{$self->{phase}}{conflicts}{$module} = $requirement || 0;
-}
-
-# Module::Install compatible shortcuts
-
-sub configure_requires {
-    my($self, @args) = @_;
-    $self->on(configure => sub { $self->requires(@args) });
-}
-
-sub build_requires {
-    my($self, @args) = @_;
-    $self->on(build => sub { $self->requires(@args) });
-}
-
-sub test_requires {
-    my($self, @args) = @_;
-    $self->on(test => sub { $self->requires(@args) });
-}
-
-sub author_requires {
-    my($self, @args) = @_;
-    $self->on(develop => sub { $self->requires(@args) });
-}
-
-package Module::CPANfile;
-
 1;
 
 __END__
@@ -239,6 +186,9 @@ Module::CPANfile - Parse cpanfile
   my $file = Module::CPANfile->load("cpanfile");
   my $prereqs = $file->prereqs; # CPAN::Meta::Prereqs object
 
+  my @features = $file->features; # CPAN::Meta::Feature objects
+  my $merged_prereqs = $file->prereqs_with(@identifiers); # CPAN::Meta::Prereqs
+
   $file->merge_meta('MYMETA.json');
 
 =head1 DESCRIPTION
@@ -285,6 +235,15 @@ Returns L<CPAN::Meta::Prereqs> object out of the parsed cpanfile.
 
 Returns a hash reference that should be passed to C<< CPAN::Meta::Prereqs->new >>.
 
+=item features
+
+Returns a list of features available in the cpanfile as L<CPAN::Meta::Feature>.
+
+=item prereqs_with(@identifiers), effective_prereqs(\@identifiers)
+
+Returns L<CPAN::Meta::Prereqs> object, with merged prereqs for
+features identified with the C<@identifiers>.
+
 =item to_string($include_empty)
 
   $file->to_string;
@@ -321,7 +280,7 @@ warnings to users is a caller's responsibility.
   $file->merge_meta('META.yml');
   $file->merge_meta('MYMETA.json', '2.0');
 
-Merge the effective prereqs with Meta speicifcation loaded from the
+Merge the effective prereqs with Meta specification loaded from the
 given META file, using CPAN::Meta. You can specify the META spec
 version in the second argument, which defaults to 1.4 in case the
 given file is YAML, and 2 if it is JSON.
@@ -336,4 +295,4 @@ Tatsuhiko Miyagawa
 
 L<cpanfile>, L<CPAN::Meta>, L<CPAN::Meta::Spec>
 
-=cut
\ No newline at end of file
+=cut
